#include <config.h>

#include <API/APICast.h>
#include <API/APIShims.h>
#include <API/JSBase.h>
#include <API/JSContextRef.h>
#include <API/JSStringRef.h>
#include <API/JSValueRef.h>
#include <API/OpaqueJSString.h>
#include <assert.h>
#include <bytecode/SamplingTool.h>
#include <debugger/Debugger.h>
#include <debugger/DebuggerCallFrame.h>
#include <parser/SourceProvider.h>
#include <profiler/Profiler.h>
#include <runtime/InitializeThreading.h>
#include <runtime/JSFunction.h>
#include <runtime/JSNotAnObject.h>
#include <runtime/JSObject.h>
#include <runtime/Structure.h>
#include <stdarg.h>
#include <wtf/HashMap.h>
#include <wtf/MainThread.h>
#include <wtf/Vector.h>
#include <wtf/unicode/UTF8.h>

// #define DEBUGGER_DEBUG

typedef void (*BreakpointCallback)(intptr_t sourceID, int line, int reason, void* userData);
enum { Breakpoint, Exception, Step };
extern "C" JSStringRef Netflix_JSStringCreate(const char* string, int length);


static JSC::SourceProvider* idToSourceProvider(intptr_t sourceID)
{
    // ### horrible hack
    return reinterpret_cast<JSC::SourceProvider*>(sourceID);
}

namespace WTF
{
    // Make JSC::UString work in a HashMap
    template<> struct HashTraits<JSC::UString> : GenericHashTraits<JSC::UString>
    {
        static const bool hasIsEmptyValueFunction = true;
        static const bool emptyValueIsZero = true;
        static void constructDeletedValue(JSC::UString& slot) { new (NotNull, &slot) JSC::UString(static_cast<char*>(0), 0); }
        static bool isDeletedValue(const JSC::UString& value) { return value.isNull(); }
        static bool isEmptyValue(const JSC::UString& str) { return str.isEmpty(); }
    };
}

struct NetflixBreakpoint
{
    JSC::UString file;
    intptr_t sourceID;
    int line;
    JSC::UString condition;

    NetflixBreakpoint(const JSC::UString& f, int l)
        : file(f), sourceID(0), line(l)
    {
    }
    NetflixBreakpoint(intptr_t s, int l)
        : sourceID(s), line(l)
    {
    }
    NetflixBreakpoint(const JSC::UString& f, int l, const JSC::UString& c)
        : file(f), sourceID(0), line(l), condition(c)
    {
    }

    bool operator==(const NetflixBreakpoint& other) const
    {
        return (sourceID == other.sourceID || file == other.file) && line == other.line;
    }
};

class NetflixStack
{
public:
    void push(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber);
    void update(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber);
    void pop();

    bool isEmpty() const { return mCallFrames.isEmpty(); }
    size_t size() const { return mCallFrames.size(); }

    void clear() { mCallFrames.clear(); mFrameNumber = 1; }

    const JSC::DebuggerCallFrame& currentFrame() const { return frameAt(mFrameNumber); }
    const JSC::DebuggerCallFrame& frameAt(size_t frame) const { return mCallFrames.at(frame - 1).frame; }
    JSC::DebuggerCallFrame& currentFrame() { return frameAt(mFrameNumber); }
    JSC::DebuggerCallFrame& frameAt(size_t frame) { return mCallFrames.at(frame - 1).frame; }

    intptr_t currentSourceID() const { return sourceIDAt(mFrameNumber); }
    intptr_t sourceIDAt(size_t frame) const { return mCallFrames.at(frame - 1).sourceID; }
    int currentLineNumber() const { return lineNumberAt(mFrameNumber); }
    int lineNumberAt(size_t frame) const { return mCallFrames.at(frame - 1).lineNumber; }

    void setAboutToStepOut(bool value) { mCallFrames.first().aboutToStepOut = value; }

    enum StepState {
        StepNone = 0x0,
        StepOver = 0x1,
        StepInto = 0x2|StepOver // step into should also be step over
    };
    int stepState() const;
    void setStepState(int s);
    void setCallerStepState(int s);
    void clearStepStates();

    void setFrameNumber(size_t frame) { if (frame > 0 && (frame == 1 || frame <= mCallFrames.size())) mFrameNumber = frame; }
    size_t frameNumber() const { return mFrameNumber; }

private:
    struct Entry
    {
        JSC::DebuggerCallFrame frame;
        intptr_t sourceID;
        int lineNumber;
        int state;
        bool aboutToStepOut;
    };
    WTF::Vector<Entry> mCallFrames;
    size_t mFrameNumber;
};

void NetflixStack::push(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
{
    Entry e = { frame, sourceID, lineNumber, StepNone, false };
    mCallFrames.prepend(e);
    mFrameNumber = 1;
}

void NetflixStack::update(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
{
    if (mCallFrames.isEmpty())
        push(frame, sourceID, lineNumber);
    else {
        Entry& e = mCallFrames.first();
        e.frame = frame;
        e.sourceID = sourceID;
        e.lineNumber = lineNumber;
    }
}

void NetflixStack::pop()
{
    mCallFrames.remove(0);
}

int NetflixStack::stepState() const
{
    WTF::Vector<Entry>::const_iterator entry = mCallFrames.begin();
    return entry->state;
}

void NetflixStack::setStepState(int s)
{
    WTF::Vector<Entry>::iterator entry = mCallFrames.begin();
    if (entry->aboutToStepOut)
        ++entry;
    entry->state = s;
}

void NetflixStack::setCallerStepState(int s)
{
    WTF::Vector<Entry>::iterator entry = mCallFrames.begin();
    if (entry->aboutToStepOut)
        ++entry;
    ++entry;
    entry->state = s;
}

void NetflixStack::clearStepStates()
{
    WTF::Vector<Entry>::iterator it = mCallFrames.begin();
    const WTF::Vector<Entry>::const_iterator end = mCallFrames.end();
    while (it != end) {
        it->state = StepNone;
        ++it;
    }
}

static inline JSC::UString quoteString(const JSC::UString& str)
{
    if (!str.impl())
        return str;
    static UChar bfrom('\\');
    static JSC::UString bto("\\\\");
    WTF::RefPtr<WTF::StringImpl> nstr = str.impl()->replace(bfrom, bto.impl());

    static UChar qfrom('"');
    static JSC::UString qto("\\\"");
    return JSC::UString(nstr.get()->replace(qfrom, qto.impl()));
}

class NetflixDebugger : public JSC::Debugger
{
public:
    WTF::HashMap<JSC::UString, JSC::SourceProvider*> providers;
    WTF::Vector<NetflixBreakpoint> breaks;
    BreakpointCallback callback;
    void* userData;
    NetflixStack callstack;
    bool inDebug;
    int attachRef;
    enum ExceptionBreak { BreakNone, BreakAll, BreakUncaught };
    ExceptionBreak breakOnException;

    static int nextId;
    struct State
    {
        WTF::Vector<NetflixBreakpoint> breaks;
        ExceptionBreak breakOnException;
    };

    static WTF::HashMap<int, State> savedState;

    struct DebugScope
    {
        DebugScope(NetflixDebugger* d) : debugger(d) { alreadyInDebug = debugger->inDebug; debugger->inDebug = true; }
        ~DebugScope() { if (!alreadyInDebug) debugger->inDebug = false; }

        NetflixDebugger* debugger;
        bool alreadyInDebug;
    };

    static NetflixDebugger* get(JSC::JSGlobalObject* global);
    static void reset(JSC::JSGlobalObject* global);

    int saveBreakpoints()
    {
        int id = ++nextId;
        if (id == -1) // avoid returning -1
            id = ++nextId;
        State state = { breaks, breakOnException };
        savedState.add(id, state);
        return id;
    }

    void restoreBreakpoints(int id)
    {
        WTF::HashMap<int, State>::iterator it = savedState.find(id);
        if (it == savedState.end())
            return;
        breaks = it->second.breaks;
        breakOnException = it->second.breakOnException;
        resolveAll(true);
        savedState.remove(it);
    }

    virtual void sourceParsed(JSC::ExecState* /*exec*/, JSC::SourceProvider* provider,
                              int /*errorLineNumber*/, const JSC::UString& /*errorMessage*/)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
        if (!provider->url().isEmpty())
            providers.add(provider->url(), provider);
        resolveAll();
        //printf("sourceParsed %p %p %d %s\n",
        //       exec, provider, errorLineNumber, quoteString(errorMessage).utf8().data());
    }

    virtual void exception(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber, bool hasHandler)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("exception\n");
#endif
        callstack.update(frame, sourceID, lineNumber);
        if (breakOnException == BreakAll || (!hasHandler && breakOnException == BreakUncaught))
            breakHere(sourceID, lineNumber, Exception);
        else
            handleBreakpoint(sourceID, lineNumber);
    }

    virtual void atStatement(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("atStatement\n");
#endif
        callstack.update(frame, sourceID, lineNumber);
        handleBreakpoint(sourceID, lineNumber);
    }

    virtual void callEvent(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("callEvent\n");
#endif
        const int state = !callstack.isEmpty() ? callstack.stepState() : 0;
        callstack.push(frame, sourceID, lineNumber);
        if (state == NetflixStack::StepInto)
            callstack.setStepState(NetflixStack::StepOver);
        handleBreakpoint(sourceID, lineNumber);
    }

    virtual void returnEvent(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("returnEvent\n");
#endif
        callstack.update(frame, sourceID, lineNumber);
        callstack.setAboutToStepOut(true);
        handleBreakpoint(sourceID, lineNumber);
        callstack.pop();
    }

    virtual void willExecuteProgram(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("willExecuteProgram\n");
#endif
        const int state = !callstack.isEmpty() ? callstack.stepState() : 0;
        callstack.push(frame, sourceID, lineNumber);
        if (state == NetflixStack::StepInto)
            callstack.setStepState(NetflixStack::StepOver);
        handleBreakpoint(sourceID, lineNumber);
    }

    virtual void didExecuteProgram(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("didExecuteProgram\n");
#endif
        callstack.update(frame, sourceID, lineNumber);
        callstack.setAboutToStepOut(true);
        handleBreakpoint(sourceID, lineNumber);
        callstack.pop();
    }

    virtual void didReachBreakpoint(const JSC::DebuggerCallFrame& frame, intptr_t sourceID, int lineNumber)
    {
        DebugScope scope(this);
        if (scope.alreadyInDebug)
            return;
#ifdef DEBUGGER_DEBUG
        printf("didReachBreakpoint\n");
#endif
        callstack.update(frame, sourceID, lineNumber);
        handleBreakpoint(sourceID, lineNumber);
    }

    void breakHere(intptr_t sourceID, int lineNumber, int reason);
    void handleBreakpoint(intptr_t sourceID, int lineNumber);
    bool hasBreakpoint(intptr_t sourceID, int lineNumber, JSC::UString** condition);
    void resolve(NetflixBreakpoint* point, bool force = false);
    void resolveAll(bool force = false);

private:
    NetflixDebugger()
        : callback(0), userData(0), inDebug(false), attachRef(0), breakOnException(BreakNone)
    {
        callstack.setFrameNumber(0);
    }

    static WTF::HashMap<JSC::JSGlobalObject*, NetflixDebugger*> debuggers;
};

WTF::HashMap<JSC::JSGlobalObject*, NetflixDebugger*> NetflixDebugger::debuggers;
int NetflixDebugger::nextId = 0;
WTF::HashMap<int, NetflixDebugger::State> NetflixDebugger::savedState;

void NetflixDebugger::reset(JSC::JSGlobalObject* global)
{
    WTF::HashMap<JSC::JSGlobalObject*, NetflixDebugger*>::iterator it = debuggers.find(global);
    if (it != debuggers.end()) {
        if (it->second->attachRef)
            it->second->detach(global);
        delete it->second;
        debuggers.remove(it);
    }
}

NetflixDebugger* NetflixDebugger::get(JSC::JSGlobalObject* global)
{
    WTF::HashMap<JSC::JSGlobalObject*, NetflixDebugger*>::const_iterator it = debuggers.find(global);
    if (it != debuggers.end())
        return it->second;
    NetflixDebugger* debugger = new NetflixDebugger;
    debuggers.add(global, debugger);
    return debugger;
}

void NetflixDebugger::resolve(NetflixBreakpoint* point, bool force)
{
    if (point->sourceID && !force)
        return;
    WTF::HashMap<JSC::UString, JSC::SourceProvider*>::const_iterator it = providers.find(point->file);
    if (it == providers.end()) {
        if (force)
            point->sourceID = 0;
        return;
    }
    point->sourceID = it->second->asID();
}

void NetflixDebugger::resolveAll(bool force)
{
    WTF::Vector<NetflixBreakpoint>::iterator it = breaks.begin();
    const WTF::Vector<NetflixBreakpoint>::const_iterator end = breaks.end();
    while (it != end) {
        if (force || !it->sourceID) {
            resolve(&(*it), force);
        }
        ++it;
    }
}

bool NetflixDebugger::hasBreakpoint(intptr_t sourceID, int line, JSC::UString** condition)
{
    NetflixBreakpoint point(sourceID, line);
    const size_t pos = breaks.find(point);
    if (pos == WTF::notFound) {
        if (condition)
            *condition = 0;
        return false;
    }
    if (condition) {
        if (!breaks.at(pos).condition.isEmpty())
            *condition = &breaks.at(pos).condition;
        else
            *condition = 0;
    }
    return true;
}

void NetflixDebugger::breakHere(intptr_t sourceID, int line, int reason)
{
    if (callback)
        callback(sourceID, line, reason, userData);
}

void NetflixDebugger::handleBreakpoint(intptr_t sourceID, int line)
{
    bool brk = false;
    JSC::UString* condition;
    int reason = Breakpoint;
    if (hasBreakpoint(sourceID, line, &condition))
        brk = true;
    else if (callstack.stepState() & NetflixStack::StepOver) {
        reason = Step;
        callstack.setStepState(NetflixStack::StepNone);
        brk = true;
    }
    if (brk && condition) {
        JSC::JSValue exc;
        JSC::JSValue val = callstack.currentFrame().evaluate(*condition, exc);
        if (!exc.isEmpty()) // a non-empty exception means we didn't hit the condition
            brk = false;
        else {
            if (val.isEmpty() || val.isNull() || val.isUndefined() || val.isFalse()
                || (val.isNumber() && val.asNumber() == 0.))
                brk = false;
        }
    }
    if (brk)
        breakHere(sourceID, line, reason);
}

class NetflixForEachCell : public JSC::MarkedBlock::VoidFunctor
{
public:
    struct Block
    {
        size_t cellSize, blockSize, blockCapacity;
        WTF::Vector<JSC::JSCell*> cells;

        Block(size_t cs, size_t bs, size_t bc)
            : cellSize(cs), blockSize(bs), blockCapacity(bc)
        {
        }
    };
    WTF::Vector<Block> blocks;

    void operator()(JSC::JSCell* cell)
    {
        blocks.last().cells.append(cell);
    }

    static size_t currentCellSize;
};

size_t NetflixForEachCell::currentCellSize = 0;

class NetflixForEachBlock : public JSC::MarkedBlock::VoidFunctor
{
public:
    NetflixForEachCell foreachCell;

    void operator()(JSC::MarkedBlock* block)
    {
        foreachCell.blocks.append(NetflixForEachCell::Block(block->cellSize(), block->size(), block->capacity()));
        block->forEachCell<NetflixForEachCell>(foreachCell);
    }
};

class JSDumper
{
public:
    static void print(const char* format, ...)
    {
        if ((data + size) - cur < BlockSize)
            allocMore();
        va_list ap;
        va_start(ap, format);
        int num = vsnprintf(cur, BlockSize, format, ap);
        if (num > BlockSize) { // retry with new size
            allocMore(num - BlockSize);
            const int newNum = vsnprintf(cur, num, format, ap);
            (void)newNum;
            assert(num == newNum);
        }
        va_end(ap);
        cur += num;
    }

    static void finalize(JSStringRef* str)
    {
        *str = Netflix_JSStringCreate(data, cur - data);
        free(data);
        data = cur = 0;
        size = 0;
    }

private:
    enum { BlockSize = 8192 };

    static void allocMore(int destSize = BlockSize)
    {
        assert(data <= cur && data + size >= cur);
        const unsigned int curpos = cur - data;
        assert(curpos <= size);
        size += destSize;
        data = static_cast<char*>(realloc(data, size));
        cur = data + curpos;
    }

    static char* data;
    static char* cur;
    static unsigned int size;
};

char* JSDumper::data = 0;
char* JSDumper::cur = 0;
unsigned int JSDumper::size = 0;

static void dumpObject(JSC::ExecState* exec, const JSC::JSObject* obj, int indent);
static void dumpValue(JSC::ExecState* exec, const JSC::JSValue& val, int indent);
static void dumpCell(JSC::ExecState* exec, const JSC::JSCell* cell, int indent);

static void dumpObject(JSC::ExecState* exec, const JSC::JSObject* obj, int indent)
{
    JSC::UString className = JSC::JSObject::className(obj);
    JSDumper::print("{\n");
    indent += 4;
    JSDumper::print("%*s\"type\": \"Object\", \"name\": \"%s\", \"id\": \"%p\",\n", indent, "", quoteString(className).utf8().data(), obj);
    JSDumper::print("%*s\"object\": {\n", indent, "");
    JSC::PropertyNameArray names(exec);
    JSC::JSObject::getPropertyNames(const_cast<JSC::JSObject*>(obj), exec, names, JSC::ExcludeDontEnumProperties);
    JSC::PropertyNameArray::const_iterator it = names.begin();
    const JSC::PropertyNameArray::const_iterator end = names.end();
    while (it != end) {
        const JSC::Identifier& ident = *it;
        indent += 4;
        JSDumper::print("%*s\"%s\": ", indent, "", quoteString(ident.ustring()).utf8().data());
        dumpValue(exec, obj->get(exec, ident), indent);
        indent -= 4;
        ++it;
        if (it != end)
            JSDumper::print(",\n");
        else
            JSDumper::print("\n");
    }
    JSDumper::print("%*s}\n", indent, "");
    indent -= 4;
    JSDumper::print("%*s}\n", indent, "");
}

static void dumpFunction(JSC::ExecState* exec, JSC::JSFunction* func, int /*indent*/)
{
    const JSC::SourceCode* code = func->sourceCode();
    if (!code || code->isNull()) {
        JSDumper::print("{ \"type\": \"Function\", \"id\": null }\n");
        return;
    }
    const JSC::UString name = func->calculatedDisplayName(exec);
    const JSC::UString file = code->provider()->url();
    JSDumper::print("{ \"type\": \"Function\", \"id\": \"%p\", \"name\": \"%s\", \"file\": \"%s\", \"start\": %d, \"end\": %d }\n",
                    func, quoteString(name).utf8().data(), quoteString(file).utf8().data(), code->startOffset(), code->endOffset());
}

static void dumpValue(JSC::ExecState* /*exec*/, const JSC::JSValue& val, int /*indent*/)
{
    if (val.isInt32())
        JSDumper::print("%d", val.asInt32());
    else if (val.isUInt32())
        JSDumper::print("%u", val.asUInt32());
    else if (val.isDouble())
        JSDumper::print("%f", val.asDouble());
    else if (val.isTrue())
        JSDumper::print("true");
    else if (val.isFalse())
        JSDumper::print("false");
    else if (val.isEmpty())
        JSDumper::print("{ \"type\": \"Empty\" }");
    else if (val.isUndefined() || val.isNull())
        // Print 'null' for undefined, JSON only allows 'null'
        JSDumper::print("null");
    else if (val.isCell() && val.asCell()->inherits(&JSC::JSFunction::s_info))
        JSDumper::print("{ \"type\": \"Function\", \"id\": \"%p\" }", val.asCell());
    else if (val.isString()) {
        JSDumper::print("{ \"type\": \"String\", \"id\": \"%p\" }", val.asCell());
    } else if (val.isObject()) {
        JSDumper::print("{ \"type\": \"Object\", \"id\": \"%p\" }", val.asCell());
    } else if (val.isCell()) {
        const JSC::JSCell* cell = val.asCell();
        JSDumper::print("{ \"type\": \"Unknown cell\", \"id\": \"%p\", \"type\": %d }",
               cell, cell->structure() ? cell->structure()->typeInfo().type() : -1);
    } else
        JSDumper::print("{ \"type\": \"Unknown value\", \"descr\": \"%s\" }", const_cast<JSC::JSValue*>(&val)->description());
}

static void dumpCell(JSC::ExecState* exec, const JSC::JSCell* cell, int indent)
{
    JSDumper::print("%*s{ \"type\": \"Cell\", \"id\": \"%p\",\n", indent, "", cell, NetflixForEachCell::currentCellSize);
    JSDumper::print("%*s\"data\": ", indent + 2, "");
    indent += 4;
    if (cell->isString()) {
        JSC::UString str = cell->getString(exec);
        JSDumper::print("{ \"type\": \"String\", \"id\": \"%p\", \"data\": \"%s\" }\n", cell, quoteString(str).utf8().data());
    } else if (cell->isObject()) {
        if (!cell->inherits(&JSC::JSNotAnObject::s_info)) {
            if (cell->inherits(&JSC::JSFunction::s_info))
                dumpFunction(exec, const_cast<JSC::JSFunction*>(static_cast<const JSC::JSFunction*>(cell)), indent);
            else
                dumpObject(exec, cell->getObject(), indent);
        }
    } else if (cell->isGetterSetter()) {
        JSDumper::print("{ \"type\": \"Getter/setter\" }\n");
    } else if (cell->isAPIValueWrapper()) {
        JSDumper::print("{ \"type\": \"API value wrapper\" }\n");
    } else {
        JSDumper::print("{ \"type\": \"Unknown cell\", \"id\": \"%p\", \"type\": %d }\n",
               cell, cell->structure() ? cell->structure()->typeInfo().type() : -1);
    }
    indent -= 4;
    JSDumper::print("%*s}\n", indent, "");
}

extern "C" {

JSValueRef Netflix_JSValueMakeObject(JSContextRef ctx, JSObjectRef object)
{
    JSC::ExecState *exec = toJS(ctx);
    return toRef(exec, toJS(exec, object));
}

bool Netflix_JSValueIsArray(JSContextRef ctx, JSValueRef value)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::JSValue val = toJS(exec, value);
    if (val.isCell()) {
        JSC::JSCell *cell = val.asCell();
        const JSC::ClassInfo *classInfo = cell->classInfo();
        while (classInfo) {
            if (!strcmp(classInfo->className, "Array")) {
                return true;
                break;
            }
            classInfo = classInfo->parentClass;
        }
    }
    return false;
}

void Netflix_DumpJSHeap(JSContextRef ctx, JSStringRef* heap)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalData &globalData = exec->globalData();
    JSC::JSGlobalObject *globalObject = exec->lexicalGlobalObject();
    if (!globalObject) {
        *heap = 0;
        return;
    }
    NetflixDebugger* debugger = NetflixDebugger::get(globalObject);
    if (debugger->attachRef) // temporarily detach the debugger
        debugger->detach(globalObject);

    NetflixForEachBlock foreach;
    globalData.heap.collectAllGarbage();
    globalData.heap.objectSpace().forEachBlock<NetflixForEachBlock>(foreach);

    size_t blockSizes = 0, blockCapacities = 0;
    WTF::Vector<NetflixForEachCell::Block>::const_iterator blockit = foreach.foreachCell.blocks.begin();
    const WTF::Vector<NetflixForEachCell::Block>::const_iterator blockend = foreach.foreachCell.blocks.end();
    JSDumper::print("{ \"data\": [\n");
    while (blockit != blockend) {
        JSDumper::print("{ \"type\": \"Block\", \"cellSize\": %u, \"blockSize\": %u, \"blockCapacity\": %u, \"cells\": [\n",
               blockit->cellSize, blockit->blockSize, blockit->blockCapacity);
        blockSizes += blockit->blockSize;
        blockCapacities += blockit->blockCapacity;
        NetflixForEachCell::currentCellSize = blockit->cellSize;
        WTF::Vector<JSC::JSCell*>::const_iterator cellit = blockit->cells.begin();
        const WTF::Vector<JSC::JSCell*>::const_iterator cellend = blockit->cells.end();
        while (cellit != cellend) {
            dumpCell(exec, *cellit, 4);
            ++cellit;
            if (cellit != cellend)
                JSDumper::print("%*s,\n", 4, "");
        }
        JSDumper::print("\n]}\n");
        ++blockit;
        if (blockit != blockend)
            JSDumper::print(",\n");
    }
    JSDumper::print("], \"totalSize\": %u, \"totalCapacity\": %u }\n", globalData.heap.size(), globalData.heap.capacity());
    {
        JSDumper::finalize(heap);
    }

    if (debugger->attachRef) // reattach the debugger if needed
        debugger->attach(globalObject);
}

void Netflix_ResetDebugger(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger::reset(global);
}

int Netflix_SaveDebuggerState(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return -1;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    return debugger->saveBreakpoints();
}

void Netflix_RestoreDebuggerState(JSContextRef ctx, int id)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    debugger->restoreBreakpoints(id);
    if (!debugger->breaks.isEmpty()) {
        if (!debugger->attachRef) {
            debugger->attach(global);
            debugger->recompileAllJSFunctions(&global->globalData());
        }
        ++debugger->attachRef;
    }
}

void Netflix_SetBreakpointCallback(JSContextRef ctx, BreakpointCallback callback, void* userData)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    debugger->callback = callback;
    debugger->userData = userData;
}

void Netflix_ResolveSourceID(intptr_t sourceID, JSStringRef* file)
{
    JSC::SourceProvider* provider = idToSourceProvider(sourceID);
    const JSC::UString url = provider->url();
    *file = JSStringCreateWithCharacters(url.characters(), url.length());
}

void Netflix_EvaluateAtBreakpoint(JSContextRef ctx, JSStringRef code, JSValueRef* ret, JSValueRef* exception)
{
    JSC::UString str(JSStringGetCharactersPtr(code), JSStringGetLength(code));
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.isEmpty())
        return;
    JSC::JSValue e;
    const JSC::DebuggerCallFrame& frame = debugger->callstack.currentFrame();
    const JSC::JSValue r = frame.evaluate(str, e);
    if (ret) {
        if (r.isEmpty() || !e.isEmpty())
            *ret = toRef(exec, JSC::jsUndefined());
        else
            *ret = toRef(exec, r);
    }
    if (!e.isEmpty() && exception) {
        *exception = toRef(exec, e); // ### exec or frame.dynamicGlobalObject()->globalExec()?
    }
}

void Netflix_AttachDebugger(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (!debugger->attachRef) {
        debugger->attach(global);
        debugger->recompileAllJSFunctions(&global->globalData());
    }
    ++debugger->attachRef;
}

void Netflix_DetachDebugger(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (!--debugger->attachRef) {
        debugger->detach(global);
        debugger->callstack.clear();
    }
}

static inline void setBreakpoint(JSContextRef ctx, NetflixBreakpoint& point, unsigned int* position)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global) {
        *position = static_cast<unsigned int>(-1);
        return;
    }
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (!debugger->attachRef) {
        assert(debugger->breaks.isEmpty());
        debugger->attach(global);
        if (!global->globalData().dynamicGlobalObject)
            debugger->recompileAllJSFunctions(&global->globalData());
        debugger->resolve(&point);
        ++debugger->attachRef;
    } else {
        if (debugger->callstack.isEmpty())
            ++debugger->attachRef;
        debugger->resolve(&point);
        const size_t pos = debugger->breaks.find(point);
        if (pos != WTF::notFound)
            debugger->breaks.remove(pos);
    }
    debugger->breaks.append(point);
    *position = debugger->breaks.size() - 1;
}

void Netflix_SetBreakpoint(JSContextRef ctx, JSStringRef file, int line, unsigned int* position)
{
    JSC::UString f(JSStringGetCharactersPtr(file), JSStringGetLength(file));
    NetflixBreakpoint point(f, line);
    setBreakpoint(ctx, point, position);
}

void Netflix_SetConditionalBreakpoint(JSContextRef ctx, unsigned int id, JSStringRef condition)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->breaks.size() <= id)
        return;
    JSC::UString c(JSStringGetCharactersPtr(condition), JSStringGetLength(condition));
    NetflixBreakpoint& point = debugger->breaks[id];
    point.condition = c;
}

void Netflix_RemoveBreakpoint(JSContextRef ctx, int id)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (id <= 0 || static_cast<unsigned int>(id) > debugger->breaks.size())
        return;
    debugger->breaks.remove(id - 1);
    if (debugger->breaks.isEmpty() && !--debugger->attachRef) {
        debugger->detach(global);
        debugger->callstack.clear();
    }
}

// duplicated in ScriptEngineJSC.cpp
struct Netflix_JSBreakpoint
{
    JSStringRef url, condition;
    int line;
};

int Netflix_ListBreakpoints(JSContextRef ctx, Netflix_JSBreakpoint *breakpoints, int max)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global) {
        return 0;
    }
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->breaks.isEmpty()) {
        return 0;
    }

    WTF::String str;
    WTF::Vector<NetflixBreakpoint>::const_iterator it = debugger->breaks.begin();
    const WTF::Vector<NetflixBreakpoint>::const_iterator end = debugger->breaks.end();
    int i = 0;
    while (it != end && i < max) {
        breakpoints[i].line = it->line;
        breakpoints[i].url = JSStringCreateWithCharacters(it->file.characters(), it->file.length());
        breakpoints[i].condition = JSStringCreateWithCharacters(it->condition.characters(), it->condition.length());
        ++it;
        ++i;
    }
    return i;
}

void Netflix_StepOver(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.isEmpty())
        return;
    debugger->callstack.setStepState(NetflixStack::StepOver);
}

void Netflix_StepInto(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.isEmpty())
        return;
    debugger->callstack.setStepState(NetflixStack::StepInto);
}

void Netflix_Continue(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.isEmpty())
        return;
    debugger->callstack.clearStepStates();
}

void Netflix_Finish(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.size() < 2)
        return;
    debugger->callstack.setStepState(NetflixStack::StepNone);
    debugger->callstack.setCallerStepState(NetflixStack::StepOver);
}

// duplicated in ScriptEngineJSC.cpp
struct Netflix_JSStackFrame
{
    JSStringRef url, function;
    int line;
};

int Netflix_StackTrace(JSContextRef ctx, Netflix_JSStackFrame *frames, int max)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return 0;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    size_t stackSize = debugger->callstack.size();
    if (static_cast<size_t>(max) < stackSize)
        stackSize = max;
    for (size_t i = 1; i <= stackSize; ++i) {
        const JSC::DebuggerCallFrame& frame = debugger->callstack.frameAt(i);
        const intptr_t sourceID = debugger->callstack.sourceIDAt(i);
        frames[i - 1].line = debugger->callstack.lineNumberAt(i);
        JSC::SourceProvider* provider = idToSourceProvider(sourceID);
        const JSC::UString fname = frame.calculatedFunctionName();
        frames[i - 1].function = JSStringCreateWithCharacters(fname.characters(), fname.length());
        const JSC::UString url = provider->url();
        frames[i - 1].url = JSStringCreateWithCharacters(url.characters(), url.length());
    }
    return stackSize;
}

void Netflix_SetCurrentFrame(JSContextRef ctx, size_t frameNo)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.size() < 2)
        return;
    debugger->callstack.setFrameNumber(frameNo);
}

size_t Netflix_GetCurrentFrame(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return 0;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.size() < 2)
        return 0;
    return debugger->callstack.frameNumber();
}

void Netflix_Identifiers(JSContextRef ctx, JSStringRef** identifiers, size_t* count)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global) {
        *count = 0;
        return;
    }
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    if (debugger->callstack.isEmpty()) {
        *count = 0;
        return;
    }
    WTF::Vector<JSC::UString> idents;
    const JSC::DebuggerCallFrame& frame = debugger->callstack.currentFrame();
    JSC::ScopeChainIterator scopeit = frame.scopeChain()->begin();
    const JSC::ScopeChainIterator scopeend = frame.scopeChain()->end();
    while (scopeit != scopeend) {
        JSC::JSObject* o = scopeit->get();
        JSC::PropertyNameArray names(global->globalExec());
        o->methodTable()->getPropertyNames(o, global->globalExec(), names, JSC::ExcludeDontEnumProperties);
        JSC::PropertyNameArray::const_iterator nameit = names.begin();
        const JSC::PropertyNameArray::const_iterator nameend = names.end();
        while (nameit != nameend) {
            idents.append(nameit->ustring());
            ++nameit;
        }
        ++scopeit;
    }
    if (idents.isEmpty()) {
        *count = 0;
        return;
    }
    *identifiers = static_cast<JSStringRef*>(malloc(sizeof(JSStringRef) * idents.size()));
    JSStringRef* cur = *identifiers;
    WTF::Vector<JSC::UString>::const_iterator identit = idents.begin();
    const WTF::Vector<JSC::UString>::const_iterator identend = idents.end();
    while (identit != identend) {
        *cur = JSStringCreateWithCharacters(identit->characters(), identit->length());
        ++cur;
        ++identit;
    }
    *count = idents.size();
}

size_t Netflix_GarbageCollect(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return 0;
    JSC::JSGlobalData& data = global->globalData();
    const size_t prev = data.heap.size();
    data.heap.collectAllGarbage();
    const size_t now = data.heap.size();
    return (prev > now) ? prev - now : 0;
}

size_t Netflix_HeapSize(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return 0;
    JSC::JSGlobalData& data = global->globalData();
    return data.heap.size();
}

void Netflix_JSFunctionCode(JSObjectRef object, JSStringRef* str)
{
    *str = 0;
    const JSC::JSObject* obj = toJS(object);
    if (!obj->inherits(&JSC::JSNotAnObject::s_info)) {
        if (obj->inherits(&JSC::JSFunction::s_info)) {
            const JSC::JSFunction* func = static_cast<const JSC::JSFunction*>(obj);
            const JSC::SourceCode* code = func->sourceCode();
            if (code) {
                const JSC::UString source = code->toString();
                *str = JSStringCreateWithCharacters(source.characters(), source.length());
            }
        }
    }
}

int Netflix_IgnoreExceptions(JSContextRef ctx, int ignore)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return -1;
    NetflixDebugger* debugger = NetflixDebugger::get(global);
    const NetflixDebugger::ExceptionBreak was = debugger->breakOnException;
    if (ignore < 0 || ignore > 3)
        return -1;
    debugger->breakOnException = static_cast<NetflixDebugger::ExceptionBreak>(ignore);
    return was;
}

void Netflix_JSCInitializeThreading()
{
    JSC::initializeThreading();
}

void Netflix_WTFInitializeMainThread()
{
    WTF::initializeMainThread();
}

JSStringRef Netflix_JSStringCreate(const char* string, int length)
{
    JSC::initializeThreading();
    if (string) {
        WTF::Vector<UChar, 1024> buffer(length);
        UChar* p = buffer.data();
        if (WTF::Unicode::convertUTF8ToUTF16(&string, string + length, &p, p + length) == WTF::Unicode::conversionOK) {
            return OpaqueJSString::create(buffer.data(), p - buffer.data()).leakRef();
        }

        p = buffer.data();
        for (int i = 0; i < length; ++i)
            p[i] = string[i];
        JSStringRef ref = JSStringCreateWithCharacters(p, length);
        return ref;
    }

    // Null string.
    return OpaqueJSString::create().leakRef();
}

JSContextGroupRef Netflix_JSContextGroupCreate()
{
    JSC::initializeThreading();
    return toRef(JSC::JSGlobalData::createContextGroup(JSC::ThreadStackTypeLarge).leakRef());
}

void Netflix_ReleaseExecutableMemory(JSContextRef ctx)
{
    JSC::ExecState *exec = toJS(ctx);
    JSC::APIEntryShim shim(exec);
    JSC::JSGlobalObject *global = exec->lexicalGlobalObject();
    if (!global)
        return;
    global->globalData().releaseExecutableMemory();
}

} // extern "C"
